@@ -1,5 +1,5 @@
 # This file is part of Xpra.
-# Copyright (C) 2010-2017 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -15,8 +15,8 @@ import signal
 from xpra.util import envint, envbool
 from xpra.os_util import POSIX
 from xpra.log import Logger
-log = Logger("server", "util", "exec")
 
+log = Logger("server", "util", "exec")
 
 USE_PROCESS_POLLING = not POSIX or envbool("XPRA_USE_PROCESS_POLLING")
 POLL_DELAY = envint("XPRA_POLL_DELAY", 2)
@@ -35,9 +35,11 @@ def reaper_cleanup():
     s = singleton
     if not s:
         return
-    s.reap()
-    s.poll()
-    singleton = None
+    singleton.cleanup()
+    #keep it around,
+    #so we don't try to reinitialize it from the wrong thread
+    #(signal requires the main thread)
+    #singleton = None
 
 
 class ProcInfo(object):
@@ -80,14 +82,6 @@ class ChildReaper(object):
             log("using process polling every %s seconds", POLL_DELAY)
             self.glib.timeout_add(POLL_DELAY*1000, self.poll)
         else:
-            #with a less buggy python, we can just check the list of pids
-            #whenever we get a SIGCHLD
-            #however.. subprocess.Popen will no longer work as expected
-            #see: http://bugs.python.org/issue9127
-            #so we must ensure certain things that exec happen first:
-            from xpra.version_util import get_platform_info
-            get_platform_info()
-
             signal.signal(signal.SIGCHLD, self.sigchld)
             # Check once after the mainloop is running, just in case the exit
             # conditions are satisfied before we even enter the main loop.
@@ -97,6 +91,12 @@ class ChildReaper(object):
                 return False # Only call once
             self.glib.timeout_add(0, check_once)
 
+    def cleanup(self):
+        self.reap()
+        self.poll()
+        self._proc_info = []
+        self._quit = None
+
     def add_process(self, process, name, command, ignore=False, forget=False, callback=None):
         pid = process.pid
         assert pid>0, "process has no pid!"
@@ -109,11 +109,11 @@ class ChildReaper(object):
         procinfo.callback = callback
         procinfo.process = process
         procinfo.returncode = process.poll()
-        procinfo.dead = False
+        procinfo.dead = procinfo.returncode is not None
         log("add_process(%s, %s, %s, %s, %s) pid=%s", process, name, command, ignore, forget, pid)
         #could have died already:
         self._proc_info.append(procinfo)
-        if procinfo.returncode is not None:
+        if procinfo.dead:
             self.add_dead_process(procinfo)
         return procinfo
 
@@ -133,10 +133,11 @@ class ChildReaper(object):
         #see if we are meant to exit-with-children
         #see if we still have procinfos alive (and not meant to be ignored)
         self.poll()
-        alive = [procinfo for procinfo in tuple(self._proc_info) if (not procinfo.ignore and not procinfo.dead)]
+        alive = tuple(procinfo for procinfo in tuple(self._proc_info)
+                      if (not procinfo.ignore and not procinfo.dead))
         cb = self._quit
         log("check() alive=%s, quit callback=%s", alive, cb)
-        if len(alive)==0:
+        if not alive:
             if cb:
                 self._quit = None
                 cb()
@@ -144,7 +145,12 @@ class ChildReaper(object):
         return True
 
     def sigchld(self, signum, frame):
-        log("sigchld(%s, %s)", signum, frame)
+        #we risk race conditions if doing anything in the signal handler,
+        #better run in the main thread asap:
+        self.glib.idle_add(self._sigchld, signum, str(frame))
+
+    def _sigchld(self, signum, frame_str):
+        log("sigchld(%s, %s)", signum, frame_str)
         self.reap()
 
     def get_proc_info(self, pid):
@@ -190,7 +196,7 @@ class ChildReaper(object):
             #forget it:
             try:
                 self._proc_info.remove(procinfo)
-            except:
+            except ValueError:
                 log("failed to remove %s from proc info list", procinfo, exc_info=True)
         log("updated procinfo=%s", procinfo)
         self.check()
@@ -200,7 +206,7 @@ class ChildReaper(object):
         while POSIX:
             log("reap() calling os.waitpid%s", (-1, "WNOHANG"))
             try:
-                pid, _ = os.waitpid(-1, os.WNOHANG)
+                pid = os.waitpid(-1, os.WNOHANG)[0]
             except OSError:
                 break
             log("reap() waitpid=%s", pid)
@@ -213,8 +219,8 @@ class ChildReaper(object):
         info = {
                 "children"  : {
                                "total"      : len(iv),
-                               "dead"       : len([x for x in iv if x.dead]),
-                               "ignored"    : len([x for x in iv if x.ignore]),
+                               "dead"       : len(tuple(True for x in iv if x.dead)),
+                               "ignored"    : len(tuple(True for x in iv if x.ignore)),
                                }
                 }
         pi = sorted(self._proc_info, key=lambda x: x.pid, reverse=True)
